/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* vim:expandtab:shiftwidth=4:tabstop=4:
 */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef __nsCommonWidget_h__
#define __nsCommonWidget_h__



class nsCommonWidget : public nsBaseWidget {
public:
    nsCommonWidget();
    virtual ~nsCommonWidget();


protected:
};

#endif /* __nsCommonWidget_h__ */
