/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef _Included_cip_sipcc_CodecMask
#define _Included_cip_sipcc_CodecMask
#ifdef __cplusplus
extern "C" {
#endif
#undef cip_sipcc_CodecMask_DSP_RESOURCE_G711
#define cip_sipcc_CodecMask_DSP_RESOURCE_G711 1L
#undef cip_sipcc_CodecMask_DSP_RESOURCE_G729A
#define cip_sipcc_CodecMask_DSP_RESOURCE_G729A 2L
#undef cip_sipcc_CodecMask_DSP_RESOURCE_G729B
#define cip_sipcc_CodecMask_DSP_RESOURCE_G729B 4L
#undef cip_sipcc_CodecMask_DSP_RESOURCE_LINEAR
#define cip_sipcc_CodecMask_DSP_RESOURCE_LINEAR 8L
#undef cip_sipcc_CodecMask_DSP_RESOURCE_G722
#define cip_sipcc_CodecMask_DSP_RESOURCE_G722 16L
#undef cip_sipcc_CodecMask_DSP_RESOURCE_iLBC
#define cip_sipcc_CodecMask_DSP_RESOURCE_iLBC 32L
#undef cip_sipcc_CodecMask_DSP_RESOURCE_iSAC
#define cip_sipcc_CodecMask_DSP_RESOURCE_iSAC 64L
#undef cip_sipcc_CodecMask_DSP_DECODEONLY
#define cip_sipcc_CodecMask_DSP_DECODEONLY 0L
#undef cip_sipcc_CodecMask_DSP_ENCODEONLY
#define cip_sipcc_CodecMask_DSP_ENCODEONLY 1L
#undef cip_sipcc_CodecMask_DSP_FULLDUPLEX
#define cip_sipcc_CodecMask_DSP_FULLDUPLEX 2L
#undef cip_sipcc_CodecMask_DSP_IGNORE
#define cip_sipcc_CodecMask_DSP_IGNORE 3L
#ifdef __cplusplus
}
#endif
#endif
