/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef _CPR_DARWIN_TST_H_
#define _CPR_DARWIN_TST_H_

#define APP_NAME "SIPCC-"
#define DEB_F_PREFIX APP_NAME"%s: %s: "
#define CPR "CPR"
#define DEB_F_PREFIX_ARGS(msg_name, func_name) msg_name, func_name


#endif
